""" This package declares meta data for SAI schema.
"""

# Force __all__ modules to be imported when importing core package
from core import *

# Only these modules will be imported with "import *"
__all__ = ["entity", "relationship"]
